''
'' Propeller Direction Cosine Matrix implementation by Jason Dorie
''
'' Based largely on the work by Premerlani and Bizard
''


CON
  FixedBits = 28           'Number of bits of precision to use internally.  DO NOT CHANGE (this works because of the mults I'm using)

  AccelBits = 8            'Number of bits of fixed point precision the accelerometer gives for 1g

  'GyroScale = 858993459    '(1<<32)/5 = 858993459 - Used to scale the gyro down by 5.  I could've used 5, but this is more tunable 

  Fixed_One = 1 << FixedBits

  Filter_Heavy =    $2AAAAAAA         '0.1666 ** 2^32  (RC = 0.0200)   
  Filter_Moderate = $2F9DB22D         '0.1860 ** 2^32  (RC = 0.0175)   
  Filter_Light =    $4923A29C         '0.2857 ** 2^32  (RC = 0.0100)   


VAR
  long Temp0, Temp1, Temp2                              ' Temporary work space
  long Temp3, Temp4, Temp5
  long Temp6, Temp7, Temp8

  long Gx, Gy, Gz                                       ' Gyro readings (raw)
  long Ax, Ay, Az                                       ' Accelerometer readings (filtered)

  long Ref0, Ref1, Ref2                                 ' Reference matrix
  long Ref3, Ref4, Ref5
  long Ref6, Ref7, Ref8                        

  long YAxisTemp0, YAxisTemp1, YAxisTemp2 
                                     
  long Correction0, Correction1, Correction2            ' Amount of drift correction to be applied
  long DownCorrect0, DownCorrect1, DownCorrect2
  long BiasCorrect0, BiasCorrect1, BiasCorrect2
  long ux_vy_wz, OM_CosTheta, CosTheta, SinTheta

  long GyroScale
  long CorrectionDivisor


PUB Init
 
  'Initialize the reference matrix to identity, fixed point format, standard identity matrix (X = left, Y = up, Z = forward)
  Ref0 := Fixed_One 
  Ref1 := 0 
  Ref2 := 0

  Ref3 := 0 
  Ref4 := Fixed_One 
  Ref5 := 0 

  Ref6 := 0 
  Ref7 := 0 
  Ref8 := Fixed_One

  Correction0 := 0
  Correction1 := 0
  Correction2 := 0   

  Gx := Gy := Gz := 0
  

PUB GetM00
  return Ref0
PUB GetM01
  return Ref1
PUB GetM02
  return Ref2
PUB GetM10
  return Ref3
PUB GetM11
  return Ref4
PUB GetM12
  return Ref5
PUB GetM20
  return Ref6
PUB GetM21
  return Ref7
PUB GetM22
  return Ref8

PUB GetCorrection0
  return Correction0

PUB GetCorrection1
  return Correction1
  
PUB GetCorrection2
  return Correction2

PUB SetValues( _Gx, _Gy, _Gz, _Ax, _Ay, _Az )

''Sets the values that the Update function uses for the Gyro and Accelerometer readings.
''The gyro readings should be raw sensor readings - use the GyroScale value to adjust the
''response of the filter until it matches the "real world".  Having a know reference speed,
''like a turntable (record player) can help greatly.

''The accelerometer readings are also raw, but adjusted so that "level" is 0 in both X and Y.
''Use the "AccelBits" value to scale the accelerometer readings such that 1g = 1.0 in fixed
''point format.  With 15 bits precision, 1.0 = $8000.  If your accelerometer requires scaling
''as opposed to shifting, change the "GetAccelAsVector" function to do so. 

  Gx := _Gx
  Gy := _Gy
  Gz := _Gz

  Ax := _Ax
  Ay := _Ay
  Az := _Az


PUB SetDownCorrect( dx, dy, dz )

  ''Sets the value used to adjust input from the accelerometer.  This value should be set to
  ''the vector seen when the board is level.  Internally (0,1,0) is subtracted to produce the
  ''final correction factor.  The DownCorrect value is subtracted from the accelerometer vector
  ''to produce a true, level, 1G value when the board is stationary and level.
  ''Note that the accelerometer reads DOWN as Z, which is switched to the Y component here.

  DownCorrect0 := -dx
  DownCorrect1 := (1<<FixedBits) - dz
  DownCorrect2 := -dy


PUB SetBiasCorrect( dx, dy, dz )

  ''Sets the value used to adjust input from the accelerometer.  This value should be set to
  ''the vector computed as the average of the min and max values for each axis of the accelerometer
  ''over a complete set of rotations.

  ''The BiasCorrect value is subtracted from the accelerometer vector
  ''to produce a true, level, 1G value when the board is stationary and level.
  ''Note that the accelerometer reads DOWN as Z, which is switched to the Y component here.

  BiasCorrect0 := dx
  BiasCorrect1 := dz
  BiasCorrect2 := dy


PUB GetMatrixValue( index )
  return long[@Ref0][index]


PUB GetAccelAsVector( Vect )
  'Note that because the gyro and accelerometer use different meanings for x,y,z, the axis get moved.  I consider a Z
  'rotation a rotation around the Z (forward) axis.
  long[Vect][0] := Ax << constant(FixedBits-AccelBits)
  long[Vect][1] := Az << constant(FixedBits-AccelBits)
  long[Vect][2] := Ay << constant(FixedBits-AccelBits)                 



PUB SetGyroScale(UpdateFreq)
  GyroScale := 1638 * 200 / UpdateFreq 


PUB Update | Gravity0, Gravity1, Gravity2, cGrav0, cGrav1, cGrav2, iGx, iGy, iGz, Len, ErrorAxis0, ErrorAxis1, ErrorAxis2

  ''This is where the magic happens - update everything.  Call SetValues before calling this function.

  'Convert the gyro readings into proper range for the DCM inputs.  Also, swap the axis
  'around and adjust signs to coincide with my frame of reference

  '1638 = (G << 13)/5     - Note the /5 is gyro scale for an ITG3200, and assumes an update rate of 200Hz 
  '1330 = (G << 13)/6.25  - Note the /6.25 is gyro scale for an ITG3200, and assumes an update rate of 250Hz 
  ' Other gyros or update rates will require different gyro scale factors                       

  '200Hz update scale factor
  'iGx := ( Gy * 1638) + Correction0                     'Scale the gyro readings, add drift correction             
  'iGy := (-Gz * 1638) + Correction1                     
  'iGz := (-Gx * 1638) + Correction2
  iGx := ( Gy * GyroScale) + Correction0                     'Scale the gyro readings, add drift correction             
  iGy := ( Gz * GyroScale) + Correction1                     
  iGz := (-Gx * GyroScale) + Correction2

  '250Hz update scale factor
  'iGx := ( Gy * 1330) + Correction0                     'Scale the gyro readings, add drift correction             
  'iGy := (-Gz * 1330) + Correction1                     
  'iGz := (-Gx * 1330) + Correction2
   

  'Rotate the current reference matrix into the temporary
  RotateRefToTempMatrix( iGx, iGy, iGz ) 

  'Normalize & orthonormalize the result
  RenormalizeTempMatrix

  'Copy the result back into the reference matrix
  longmove( @Ref0, @Temp0, 9 )  
   
   
  'Gravity (or some approximation thereof) is the accelerometer
  'GetAccelAsVector( @Gravity )  'Swaps the axis, scales the values - inline saves 1168 cycles

  'Format the gravity vector, subtract the BiasCorrect adjustment (accelerometer reading bias)  
  Gravity0 := (Ax << constant(FixedBits-AccelBits)) - BiasCorrect0
  Gravity1 := (Az << constant(FixedBits-AccelBits)) - BiasCorrect1
  Gravity2 := (Ay << constant(FixedBits-AccelBits)) - BiasCorrect2                 


  'GravityNorm = Gravity.Normalized();
  'NormalizeVector( @Gravity0 )

  'Faster normalization - do it inline
  Len := DotVector( @Gravity0, @Gravity0 ) 
  Len := (constant(3 << FixedBits) - Len) ~> 1
  Gravity0  := (Gravity0 ** Len) << 4
  Gravity1  := (Gravity1 ** Len) << 4
  Gravity2  := (Gravity2 ** Len) << 4


  'FixedVector WorldUp = new FixedVector(BodyEstCorr.m[1, 0], BodyEstCorr.m[1, 1], BodyEstCorr.m[1, 2]);
  YAxisTemp0 := Ref3
  YAxisTemp1 := Ref4
  YAxisTemp2 := Ref5 
  
  'ErrorAxis = GravityNorm.Cross(WorldUp)
  CrossVector( @Gravity0, @YAxisTemp0, @ErrorAxis0 )    

  'Scale the error value down enough to feed back into the update loop.  Should I convert this to a PI loop?
  NormalizeVector( @ErrorAxis0 )

  'Scale down the Error term and store it in the Correction vector
  Correction0 := ErrorAxis0 / CorrectionDivisor
  Correction1 := ErrorAxis1 / CorrectionDivisor
  Correction2 := ErrorAxis2 / CorrectionDivisor


PUB SetCorrectionDivisor(val)
  CorrectionDivisor := val


PUB RotateRefToTempMatrix( Rx, Ry, Rz )

  { Shown here for clarity - The F1 terms (fixed point one) are removed from the real code for speed
    F1 := Fixed_One  
     
    long[Dest][0] := (long[Mat][0] *  F1 + long[Mat][1] *  Rz + long[Mat][2] * -Ry) ~> FixedBits
    long[Dest][1] := (long[Mat][0] * -Rz + long[Mat][1] *  F1 + long[Mat][2] *  Rx) ~> FixedBits
    long[Dest][2] := (long[Mat][0] *  Ry + long[Mat][1] * -Rx + long[Mat][2] *  F1) ~> FixedBits
     
    long[Dest][3] := (long[Mat][3] *  F1 + long[Mat][4] *  Rz + long[Mat][5] * -Ry) ~> FixedBits
    long[Dest][4] := (long[Mat][3] * -Rz + long[Mat][4] *  F1 + long[Mat][5] *  Rx) ~> FixedBits
    long[Dest][5] := (long[Mat][3] *  Ry + long[Mat][4] * -Rx + long[Mat][5] *  F1) ~> FixedBits
     
    long[Dest][6] := (long[Mat][6] *  F1 + long[Mat][7] *  Rz + long[Mat][8] * -Ry) ~> FixedBits
    long[Dest][7] := (long[Mat][6] * -Rz + long[Mat][7] *  F1 + long[Mat][8] *  Rx) ~> FixedBits
    long[Dest][8] := (long[Mat][6] *  Ry + long[Mat][7] * -Rx + long[Mat][8] *  F1) ~> FixedBits
    }

  Temp0 := Ref0 + ((Ref1 **  Rz + Ref2 ** -Ry) << 4)
  Temp1 := Ref1 + ((Ref0 ** -Rz + Ref2 **  Rx) << 4)
  Temp2 := Ref2 + ((Ref0 **  Ry + Ref1 ** -Rx) << 4)
  
  Temp3 := Ref3 + ((Ref4 **  Rz + Ref5 ** -Ry) << 4)
  Temp4 := Ref4 + ((Ref3 ** -Rz + Ref5 **  Rx) << 4)
  Temp5 := Ref5 + ((Ref3 **  Ry + Ref4 ** -Rx) << 4)

  Temp6 := Ref6 + ((Ref7 **  Rz + Ref8 ** -Ry) << 4)
  Temp7 := Ref7 + ((Ref6 ** -Rz + Ref8 **  Rx) << 4)
  Temp8 := Ref8 + ((Ref6 **  Ry + Ref7 ** -Rx) << 4)


PUB RenormalizeTempMatrix | OrthoError, XErr0, XErr1, XErr2

  'FixedVector xRow = new FixedVector(m[0, 0], m[0, 1], m[0, 2]);
  'FixedVector yRow = new FixedVector(m[1, 0], m[1, 1], m[1, 2]);
   
  'OrthoError := -DotVector( @Temp0 , @Temp0+12 ) / 2            'Will be zero if there's no error, and scale by half
  OrthoError := (-((Temp0 ** Temp3) + (Temp1 ** Temp4) + (Temp2 ** Temp5)) << 4) / 2

   
  'Scale rows x and y by the error
  'FixedVector xErr = xRow.Scale( error );
  XErr0 := (Temp0 ** OrthoError) << 4
  XErr1 := (Temp1 ** OrthoError) << 4
  XErr2 := (Temp2 ** OrthoError) << 4

  'FixedVector yErr = yRow.Scale( error );
  Temp0 += (Temp3 ** OrthoError) << 4
  Temp1 += (Temp4 ** OrthoError) << 4
  Temp2 += (Temp5 ** OrthoError) << 4
   
  'Add the error corrections to the opposite row vector
  'xRow = xRow.Add(yErr)
  'Temp0 += YErr0   
  'Temp1 += YErr1
  'Temp2 += YErr2
  
  'yRow = yRow.Add(xErr)
  Temp3 += XErr0   
  Temp4 += XErr1
  Temp5 += XErr2

  'Normalize the X and Y rows   
  'xRow = xRow.Normalized()
  NormalizeVector( @Temp0 )

  'yRow = yRow.Normalized()
  NormalizeVector( @Temp0 + 12 )
   
  'FixedVector zRow = xRow.Cross( yRow );
  'CrossVector( @Temp+0 , @Temp+12 , @Temp+24 ) 

  Temp6 := (Temp1 ** Temp5 - Temp2 ** Temp4) << 4
  Temp7 := (Temp2 ** Temp3 - Temp0 ** Temp5) << 4
  Temp8 := (Temp0 ** Temp4 - Temp1 ** Temp3) << 4


'In the following code I use divide instead of shift right in some cases because it's more accurate for negative
'numbers.  -1 arithmetic shift right 1 is still -1, which isn't true for positive numbers.  Divide corrects that
'even though it's a bit slower.

PUB NormalizeVector( Vect ) | Scale , SquareLen

  SquareLen := DotVector( Vect, Vect ) 
  Scale := (constant(3 << FixedBits) - SquareLen) ~> 1
  long[Vect]    := (long[Vect]    ** Scale) << 4
  long[Vect][1] := (long[Vect][1] ** Scale) << 4
  long[Vect][2] := (long[Vect][2] ** Scale) << 4


PUB DotVector( V1, V2 )
  result := ((long[V1] ** long[V2]) + (long[V1][1] ** long[V2][1]) + (long[V1][2] ** long[V2][2])) << 4


PUB CrossVector( V1, V2, Dest )

  'res.v[0] = (lv[1] * rv[2] - lv[2] * rv[1]) >> FixedMatrix.Bits;
  'res.v[1] = (lv[2] * rv[0] - lv[0] * rv[2]) >> FixedMatrix.Bits;
  'res.v[2] = (lv[0] * rv[1] - lv[1] * rv[0]) >> FixedMatrix.Bits;

  long[Dest]    := (long[V1][1] ** long[V2][2] - long[V1][2] ** long[V2][1]) << 4
  long[Dest][1] := (long[V1][2] ** long[V2]    - long[V1]    ** long[V2][2]) << 4
  long[Dest][2] := (long[V1]    ** long[V2][1] - long[V1][1] ** long[V2]   ) << 4

  